import re
import logging

from xfutils.audit_constant import *
from django.conf import settings
from django.core.cache import cache
from xfutils.utils import api_call
from .constants import CREATE_EVENTS, UPDATE_EVENTS, DELETE_EVENTS, EXCHANGE_NAME, ROUTING_KEY_CHANGEEVENT, PB_EVENTS
from .utils import generate_interservice_jwt_token

logger = logging.getLogger(__name__)

def get_diff_record(old_value, new_value):
    diff_record = {}
    diff_record['OLD_DATA'] = old_value
    diff_record['NEW_DATA'] = new_value
    return diff_record


def not_equal(new_value, old_value):
    return (new_value is None and old_value is not None) or (new_value is not None and new_value != old_value)


def get_map_diff(old_map, new_map, exclude_keys=None, diff_keys=None, mandatory_keys=None):
    diff = {}

    # Handle mandatory keys
    if mandatory_keys:
        for key in mandatory_keys:
            new_value = new_map.get(key)
            old_value = old_map.get(key)
            diff[key] = get_diff_record(old_value, new_value)

    # Handle diff keys
    if diff_keys:
        for key in diff_keys:
            new_value = new_map.get(key)
            old_value = old_map.get(key)
            diff[key] = get_diff_record(old_value, new_value)
            if not_equal(new_value, old_value):
                diff[key] = get_diff_record(old_value, new_value)

    # Handle remaining keys
    else:
        for key, new_value in new_map.items():
            if exclude_keys and key in exclude_keys:
                continue
            old_value = old_map.get(key)
            if not_equal(new_value, old_value):
                diff[key] = get_diff_record(old_value, new_value)

    return diff


def convert_list_of_maps_to_map_of_maps(list_of_maps, unique_key):
    return {m[unique_key]: m for m in list_of_maps}


def get_list_of_maps_diff(old_list_of_maps, new_list_of_maps, unique_key, exclude_keys, diff_keys, mandatory_keys):
    # Initialize list_of_maps_diff as a dictionary
    list_of_maps_diff = {}

    # Default empty list if old_list_of_maps or new_list_of_maps are None
    if not old_list_of_maps:
        old_list_of_maps = []
    if not new_list_of_maps:
        new_list_of_maps = []

    # Find common maps in both lists
    common_maps = [m for m in old_list_of_maps if m in new_list_of_maps]

    # Remove common maps from both lists
    old_list_of_maps = [m for m in old_list_of_maps if m not in common_maps]
    new_list_of_maps = [m for m in new_list_of_maps if m not in common_maps]

    # Convert lists of maps to maps of maps using unique_key
    old_map_of_maps = convert_list_of_maps_to_map_of_maps(old_list_of_maps, unique_key)
    new_map_of_maps = convert_list_of_maps_to_map_of_maps(new_list_of_maps, unique_key)

    # Initialize lists for maps that were created, updated, and deleted
    maps_created = []
    maps_deleted = []
    maps_updated = []

    # Iterate over newMapOfMaps to find updated and created maps
    for key, new_map in new_map_of_maps.items():
        if key in old_map_of_maps:
            old_map = old_map_of_maps[key]
            diff = get_map_diff(old_map, new_map, exclude_keys, diff_keys, mandatory_keys)
            maps_updated.append(diff)
        else:
            maps_created.append(new_map)

    # Iterate over oldMapOfMaps to find deleted maps
    for key, old_map in old_map_of_maps.items():
        if key not in new_map_of_maps:
            maps_deleted.append(old_map)

    # Store results in list_of_maps_diff dictionary
    list_of_maps_diff["CREATE"] = maps_created
    list_of_maps_diff["UPDATE"] = maps_updated
    list_of_maps_diff["DELETE"] = maps_deleted

    return list_of_maps_diff



def extract_jinja_variables(template_string):
    # Regex to match Jinja2 variables {{ ... }}
    pattern = r"\{\{\s*([^\}]+?)\s*\}\}"
    matches = re.findall(pattern, template_string)
    all_variables = []

    # Extract variable names and clean them up
    variables = [match.replace(" ", "").strip() for match in matches]
    for variable in variables:
        all_variables.extend( re.split(r"[ ,;|]+", variable))
    return all_variables


def load_mmd(tenantid, tenantuuid):
    models = []
    response_data = []
    platform_url = 'https://' + settings.APPMANAGER_HOST + '/api/appmanager/platform-manifest/'
    microapps_url = 'https://' + settings.APPMANAGER_HOST + '/api/appmanager/microapps-manifest/'
    token = generate_interservice_jwt_token(application_config=settings.ALL_CONFIG, tenantid=tenantid, tenantuuid=tenantuuid)
    headers = {
        'Content-Type': 'application/json',
        'Authorization': f"Bearer {token}"
    }
    response = api_call(platform_url, method='GET', headers=headers, verify=False)
    response_data.extend(response['hydra:member'])
    response = api_call(microapps_url, method='GET', headers=headers, verify=False)
    response_data.extend(response['hydra:member'])
    for data in response_data:
        model_details = data.get('manifest')
        models.extend(model_details.get('models'))
    return models


def get_mmd(tenantid, tenantuuid):
    mmd_details = cache.get('mmd_details')
    if not mmd_details:
        mmd_details = load_mmd(tenantid=tenantid, tenantuuid=tenantuuid)
        # timeout is 3hrs in seconds
        cache.set('mmd_details', mmd_details, timeout=10800 )
    return mmd_details


def extract_sub_dict(data, dotted_string):
        keys = dotted_string.split('.')
        current = data
        for key in keys:
            if isinstance(current, dict) and key in current:
                current = current[key]
            else:
                return None
        # Rebuild the path as a nested dictionary structure
        result = current
        for key in reversed(keys):
            result = {key: result}
        return result


def get_data(variables, data):
    extract_data = {}
    for variable in variables:
        if '.' in variable:
            extract_data.update(extract_sub_dict(data, variable))
        else:
            extract_data[variable] = data.get(variable, '')
    return extract_data


def get_metadata(entity_type, data, tenantid, tenantuuid):
    if not data:
        return {}
    try:
        metadata_fields = ['displayName']
        mmd_details = get_mmd(tenantid=tenantid, tenantuuid=tenantuuid)
        extract_variables = []
        for mmd in mmd_details:
            if mmd.get('type') == entity_type:
                for field in metadata_fields:
                    value = mmd.get(field)
                    extract_variables = extract_jinja_variables(value)
        return get_data(extract_variables, data)
    except Exception as e:
        logger.warn("Error finding the metadata Error %s"%str(e))
    return {}


def get_data_key(operation):
    if operation in CREATE_EVENTS + UPDATE_EVENTS:
        return 'newData'
    else:
        return 'oldData'
